--
library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use IEEE.std_logic_unsigned.all;

entity AY38910 is
	port (
		BC1			: in  std_logic;
		BDIR		: in  std_logic;
		A9N			: in  std_logic;
		DAI			: in  std_logic_vector(7 downto 0);
		IA			: in  std_logic_vector(7 downto 0);
		IB			: in  std_logic_vector(7 downto 0);
		CLK			: in  std_logic;
		RSTN		: in  std_logic;
		SNDOUT		: out std_logic;
		DAO			: out std_logic_vector(7 downto 0);
		OA			: out std_logic_vector(7 downto 0);
		OB			: out std_logic_vector(7 downto 0);
		ENBAN		: out std_logic;
		ENBBN		: out std_logic
	);
end AY38910;

architecture RTL of AY38910 is

	signal reg_tonea	: std_logic_vector(11 downto 0);
	signal reg_toneb	: std_logic_vector(11 downto 0);
	signal reg_tonec	: std_logic_vector(11 downto 0);
	signal reg_noise	: std_logic_vector(4 downto 0);
	signal reg_mixer	: std_logic_vector(7 downto 0);
	signal reg_vola		: std_logic_vector(4 downto 0);
	signal reg_volb		: std_logic_vector(4 downto 0);
	signal reg_volc		: std_logic_vector(4 downto 0);
	signal reg_envelope	: std_logic_vector(15 downto 0);
	signal reg_envsharp	: std_logic_vector(3 downto 0);
	signal reg_ioa		: std_logic_vector(7 downto 0);
	signal reg_iob		: std_logic_vector(7 downto 0);

	signal port_ia		: std_logic_vector(7 downto 0);
	signal port_ib		: std_logic_vector(7 downto 0);
	signal port_ia_f1	: std_logic_vector(7 downto 0);
	signal port_ib_f1	: std_logic_vector(7 downto 0);

	signal regadd	: std_logic_vector(7 downto 0);

	signal dao_i	: std_logic_vector(7 downto 0);

	signal dai_f1	: std_logic_vector(7 downto 0);
	signal dai_f2	: std_logic_vector(7 downto 0);
	signal dai_f3	: std_logic_vector(7 downto 0);
	signal dai_f4	: std_logic_vector(7 downto 0);

	signal bc1_f1	: std_logic;
	signal bc1_f2	: std_logic;
	signal bdir_f1	: std_logic;
	signal bdir_f2	: std_logic;
	signal a9n_f1	: std_logic;
	signal a9n_f2	: std_logic;

	signal envtrg	: std_logic;

	signal latad	: std_logic;
	signal latad_f1	: std_logic;

	signal tcntlow	: std_logic_vector(6 downto 0);

	signal tcntcha	: std_logic_vector(11 downto 0);
	signal tcntchb	: std_logic_vector(11 downto 0);
	signal tcntchc	: std_logic_vector(11 downto 0);
	signal tonecha	: std_logic;
	signal tonechb	: std_logic;
	signal tonechc	: std_logic;

	signal ncnt		: std_logic_vector(4 downto 0);
	signal ngen		: std_logic_vector(16 downto 0);

	signal envcnt	: std_logic_vector(15 downto 0);
	signal attack	: std_logic;
	signal initenv	: std_logic;
	signal holdenv	: std_logic;

	signal volenv	: std_logic_vector(3 downto 0);
	signal vola		: std_logic_vector(3 downto 0);
	signal volb		: std_logic_vector(3 downto 0);
	signal volc		: std_logic_vector(3 downto 0);

	signal ch_a		: std_logic_vector(1 downto 0);
	signal ch_b		: std_logic_vector(1 downto 0);
	signal ch_c		: std_logic_vector(1 downto 0);

	signal dsndouta	: std_logic_vector(7 downto 0);
	signal dsndoutb	: std_logic_vector(7 downto 0);
	signal dsndoutc	: std_logic_vector(7 downto 0);

	signal dsndout	: std_logic_vector(9 downto 0);
	signal dsndout2	: std_logic_vector(8 downto 0);
	signal sigma	: std_logic_vector(10 downto 0);
	signal sndout_i	: std_logic;

begin

-- BC1/BDIR (BC2 = '1')
-- 00 INACTIVE
-- 10 READ FROM PSG
-- 01 WRITE TO PSG
-- 11 LATCH ADDRESS


-- input latch
	process (CLK,RSTN)
	begin
		if (RSTN = '0') then
			bc1_f1  <= '0';
			bc1_f2  <= '0';
			bdir_f1 <= '0';
			bdir_f2 <= '0';
			a9n_f1  <= '1';
			a9n_f2  <= '1';
			dai_f1  <= (others => '0');
			dai_f2  <= (others => '0');
			dai_f3  <= (others => '0');
			dai_f4  <= (others => '0');
		elsif (CLK'event and CLK = '1') then
			bc1_f1  <= BC1;
			bc1_f2  <= bc1_f1;
			bdir_f1 <= BDIR;
			bdir_f2 <= bdir_f1;
			a9n_f1  <= A9N;
			a9n_f2  <= a9n_f1;
			dai_f1  <= DAI;
			dai_f2  <= dai_f1;
			dai_f3  <= dai_f2;
			dai_f4  <= dai_f3;
		end if;
	end process;

	latad <= '1' when (a9n_f2 = '0' and bc1_f2 = '1' and bdir_f2 = '1') else '0';

	process (CLK,RSTN)
	begin
		if (RSTN = '0') then
			latad_f1 <= '0';
		elsif (CLK'event and CLK = '1') then
			latad_f1 <= latad;
		end if;
	end process;

-- register address
	process (CLK,RSTN)
	begin
		if (RSTN = '0') then
			regadd <= (others => '0');
		elsif (CLK'event and CLK = '1') then
			if (latad = '1' and latad_f1 = '0') then
				regadd <= dai_f4;
			end if;
		end if;
	end process;

-- write register
	process (CLK,RSTN)
	begin
		if (RSTN = '0') then
			reg_tonea    <= (others => '0');
			reg_toneb    <= (others => '0');
			reg_tonec    <= (others => '0');
			reg_noise    <= (others => '0');
			reg_mixer    <= X"38";
			reg_vola     <= (others => '0');
			reg_volb     <= (others => '0');
			reg_volc     <= (others => '0');
			reg_envelope <= (others => '0');
			reg_envsharp <= (others => '0');
			reg_ioa      <= (others => '1');
			reg_iob      <= (others => '1');
			envtrg       <= '0';
		elsif (CLK'event and CLK = '1') then
			if (a9n_f2 = '0' and bc1_f2 = '0' and bdir_f2 = '1') then
				case regadd is
					when X"00" => reg_tonea( 7 downto 0)    <= dai_f4;
					when X"01" => reg_tonea(11 downto 8)    <= dai_f4(3 downto 0);
					when X"02" => reg_toneb( 7 downto 0)    <= dai_f4;
					when X"03" => reg_toneb(11 downto 8)    <= dai_f4(3 downto 0);
					when X"04" => reg_tonec( 7 downto 0)    <= dai_f4;
					when X"05" => reg_tonec(11 downto 8)    <= dai_f4(3 downto 0);
					when X"06" => reg_noise                 <= dai_f4(4 downto 0);
					when X"07" => reg_mixer                 <= dai_f4;
					when X"08" => reg_vola                  <= dai_f4(4 downto 0);
					when X"09" => reg_volb                  <= dai_f4(4 downto 0);
					when X"0A" => reg_volc                  <= dai_f4(4 downto 0);
					when X"0B" => reg_envelope( 7 downto 0) <= dai_f4;
					when X"0C" => reg_envelope(15 downto 8) <= dai_f4;
					when X"0D" => reg_envsharp              <= dai_f4(3 downto 0);
					when X"0E" => reg_ioa                   <= dai_f4;
					when X"0F" => reg_iob                   <= dai_f4;
					when others => null;
				end case;
			end if;
			if (a9n_f2 = '0' and bc1_f2 = '0' and bdir_f2 = '1' and regadd = X"0D") then
				envtrg <= '1';
			else
				envtrg <= '0';
			end if;
		end if;
	end process;

-- read register
	process (CLK,RSTN)
	begin
		if (RSTN = '0') then
			dao_i <= (others => '0');
		elsif (CLK'event and CLK = '1') then
			if (a9n_f2 = '0' and bc1_f2 = '1' and bdir_f2 = '0') then
				case regadd is
					when X"00" => dao_i <=          reg_tonea( 7 downto 0);
					when X"01" => dao_i <= "0000" & reg_tonea(11 downto 8);
					when X"02" => dao_i <=          reg_toneb( 7 downto 0);
					when X"03" => dao_i <= "0000" & reg_toneb(11 downto 8);
					when X"04" => dao_i <=          reg_tonec( 7 downto 0);
					when X"05" => dao_i <= "0000" & reg_tonec(11 downto 8);
					when X"06" => dao_i <= "000"  & reg_noise;
					when X"07" => dao_i <=          reg_mixer;
					when X"08" => dao_i <= "000"  & reg_vola;
					when X"09" => dao_i <= "000"  & reg_volb;
					when X"0A" => dao_i <= "000"  & reg_volc;
					when X"0B" => dao_i <=          reg_envelope( 7 downto 0);
					when X"0C" => dao_i <=          reg_envelope(15 downto 8);
					when X"0D" => dao_i <= "0000" & reg_envsharp;
					when X"0E" => dao_i <=          port_ia_f1;
					when X"0F" => dao_i <=          port_ib_f1;
					when others => dao_i <= regadd;
				end case;
			end if;
		end if;
	end process;


-- tone/noise counter low (4MHz)
	process (CLK,RSTN)
	begin
		if (RSTN = '0') then
			tcntlow <= (others => '0');
		elsif (CLK'event and CLK = '1') then
			tcntlow <= tcntlow + 1;
		end if;
	end process;

-- tone counter
	process (CLK,RSTN)
	begin
		if (RSTN = '0') then
			tcntcha <= (others => '0');
			tcntchb <= (others => '0');
			tcntchc <= (others => '0');
			tonecha <= '0';
			tonechb <= '0';
			tonechc <= '0';
		elsif (CLK'event and CLK = '1') then
			if (tcntlow(5 downto 0) = "111111") then

				if (tcntcha = 0) then
					if (reg_tonea = 0) then
						tcntcha <= (others => '0');
					else
						tcntcha <= reg_tonea - 1;
					end if;
					tonecha <= not tonecha;
				else
					tcntcha <= tcntcha - 1;
				end if;

				if (tcntchb = 0) then
					if (reg_toneb = 0) then
						tcntchb <= (others => '0');
					else
						tcntchb <= reg_toneb - 1;
					end if;
					tonechb <= not tonechb;
				else
					tcntchb <= tcntchb - 1;
				end if;

				if (tcntchc = 0) then
					if (reg_tonec = 0) then
						tcntchc <= (others => '0');
					else
						tcntchc <= reg_tonec - 1;
					end if;
					tonechc <= not tonechc;
				else
					tcntchc <= tcntchc - 1;
				end if;

			end if;
		end if;
	end process;

-- noise counter
	process (CLK,RSTN)
	begin
		if (RSTN = '0') then
			ncnt <= (others => '0');
			ngen <= (others => '0');
		elsif (CLK'event and CLK = '1') then
			if (tcntlow = "1111111") then
				if (ncnt = 0) then
					if (reg_tonea = 0) then
						ncnt <= (others => '0');
					else
						ncnt <= reg_noise - 1;
					end if;
					ngen <= ngen(15 downto 0) & (ngen(16) xor ngen(13) xor '1');
				else
					ncnt <= ncnt - 1;
				end if;
			end if;
		end if;
	end process;


-- envelope generator
	process (CLK,RSTN)
	begin
		if (RSTN = '0') then
			volenv  <= (others => '0');
			envcnt  <= (others => '0');
			attack  <= '0';
			initenv <= '0';
			holdenv <= '0';
		elsif (CLK'event and CLK = '1') then
			if (envtrg = '1') then
				initenv <= '1';
				holdenv <= '0';
				volenv  <= (others => '0');
			elsif (tcntlow = "1111111") then
				if (initenv = '1') then
					initenv <= '0';
					holdenv <= '0';
					if (reg_envelope = 0) then
						envcnt <= (others => '0');
					else
						envcnt <= reg_envelope - 1;
					end if;
					if (reg_envsharp(2) = '0') then
						volenv  <= "1111";
						attack  <= '0';
					else
						volenv  <= "0000";
						attack  <= '1';
					end if;

				elsif (envcnt = 0) then
					if (reg_envelope = 0) then
						envcnt <= (others => '0');
					else
						envcnt <= reg_envelope - 1;
					end if;

					if (holdenv = '0') then
						if ( (attack = '0' and volenv = "0000") or (attack = '1' and volenv = "1111") ) then

							if (reg_envsharp(3) = '0') then
								volenv <= "0000";
								holdenv <= '1';
							else
								if ( (reg_envsharp(1) xor reg_envsharp(0)) = '0') then
									volenv <= not volenv;
								else
									attack <= not attack;
								end if;
								if (reg_envsharp(0) = '1') then
									holdenv <= '1';
								end if;
							end if;

						elsif (attack = '0') then
							volenv <= volenv - 1;
						else
							volenv <= volenv + 1;
						end if;

					end if;

				else
					envcnt <= envcnt - 1;
				end if;
			end if;
		end if;
	end process;

	vola    <= volenv when (reg_vola(4) = '1') else reg_vola(3 downto 0);
	volb    <= volenv when (reg_volb(4) = '1') else reg_volb(3 downto 0);
	volc    <= volenv when (reg_volc(4) = '1') else reg_volc(3 downto 0);

-- channal mix
	ch_a <=	"10" when (reg_mixer(0) = '0' and reg_mixer(3) = '0' and tonecha = '1' and ngen(0) = '1') else
			"01" when (reg_mixer(0) = '0' and reg_mixer(3) = '0' and tonecha = '1' and ngen(0) = '0') else
			"01" when (reg_mixer(0) = '0' and reg_mixer(3) = '0' and tonecha = '0' and ngen(0) = '1') else
			"00" when (reg_mixer(0) = '0' and reg_mixer(3) = '0' and tonecha = '0' and ngen(0) = '0') else
			"10" when (reg_mixer(0) = '0' and reg_mixer(3) = '1' and tonecha = '1') else
			"00" when (reg_mixer(0) = '0' and reg_mixer(3) = '1' and tonecha = '0') else
			"10" when (reg_mixer(0) = '1' and reg_mixer(3) = '0' and ngen(0) = '1') else
			"00" when (reg_mixer(0) = '1' and reg_mixer(3) = '0' and ngen(0) = '0') else
			"00";

	ch_b <=	"10" when (reg_mixer(1) = '0' and reg_mixer(4) = '0' and tonechb = '1' and ngen(0) = '1') else
			"01" when (reg_mixer(1) = '0' and reg_mixer(4) = '0' and tonechb = '1' and ngen(0) = '0') else
			"01" when (reg_mixer(1) = '0' and reg_mixer(4) = '0' and tonechb = '0' and ngen(0) = '1') else
			"00" when (reg_mixer(1) = '0' and reg_mixer(4) = '0' and tonechb = '0' and ngen(0) = '0') else
			"10" when (reg_mixer(1) = '0' and reg_mixer(4) = '1' and tonechb = '1') else
			"00" when (reg_mixer(1) = '0' and reg_mixer(4) = '1' and tonechb = '0') else
			"10" when (reg_mixer(1) = '1' and reg_mixer(4) = '0' and ngen(0) = '1') else
			"00" when (reg_mixer(1) = '1' and reg_mixer(4) = '0' and ngen(0) = '0') else
			"00";

	ch_c <=	"10" when (reg_mixer(2) = '0' and reg_mixer(5) = '0' and tonechc = '1' and ngen(0) = '1') else
			"01" when (reg_mixer(2) = '0' and reg_mixer(5) = '0' and tonechc = '1' and ngen(0) = '0') else
			"01" when (reg_mixer(2) = '0' and reg_mixer(5) = '0' and tonechc = '0' and ngen(0) = '1') else
			"00" when (reg_mixer(2) = '0' and reg_mixer(5) = '0' and tonechc = '0' and ngen(0) = '0') else
			"10" when (reg_mixer(2) = '0' and reg_mixer(5) = '1' and tonechc = '1') else
			"00" when (reg_mixer(2) = '0' and reg_mixer(5) = '1' and tonechc = '0') else
			"10" when (reg_mixer(2) = '1' and reg_mixer(5) = '0' and ngen(0) = '1') else
			"00" when (reg_mixer(2) = '1' and reg_mixer(5) = '0' and ngen(0) = '0') else
			"00";

-- mixer
	process (CLK,RSTN)
	begin
		if (RSTN = '0') then
			dsndouta <= X"00";
			dsndoutb <= X"00";
			dsndoutc <= X"00";
		elsif (CLK'event and CLK = '1') then

			if (ch_a = "10") then
				case vola is
					when X"0" =>   dsndouta <= X"00";
					when X"1" =>   dsndouta <= X"01";
					when X"2" =>   dsndouta <= X"02";
					when X"3" =>   dsndouta <= X"03";
					when X"4" =>   dsndouta <= X"05";
					when X"5" =>   dsndouta <= X"07";
					when X"6" =>   dsndouta <= X"0A";
					when X"7" =>   dsndouta <= X"0F";
					when X"8" =>   dsndouta <= X"16";
					when X"9" =>   dsndouta <= X"1F";
					when X"A" =>   dsndouta <= X"2C";
					when X"B" =>   dsndouta <= X"3F";
					when X"C" =>   dsndouta <= X"5A";
					when X"D" =>   dsndouta <= X"7F";
					when X"E" =>   dsndouta <= X"B4";
					when X"F" =>   dsndouta <= X"FF";
					when others => dsndouta <= X"00";
				end case;
			elsif (ch_a = "01") then
				case vola is
					when X"0" =>   dsndouta <= X"00";
					when X"1" =>   dsndouta <= X"01";
					when X"2" =>   dsndouta <= X"01";
					when X"3" =>   dsndouta <= X"02";
					when X"4" =>   dsndouta <= X"03";
					when X"5" =>   dsndouta <= X"04";
					when X"6" =>   dsndouta <= X"06";
					when X"7" =>   dsndouta <= X"08";
					when X"8" =>   dsndouta <= X"0B";
					when X"9" =>   dsndouta <= X"10";
					when X"A" =>   dsndouta <= X"17";
					when X"B" =>   dsndouta <= X"20";
					when X"C" =>   dsndouta <= X"2D";
					when X"D" =>   dsndouta <= X"40";
					when X"E" =>   dsndouta <= X"5B";
					when X"F" =>   dsndouta <= X"FF";
					when others => dsndouta <= X"00";
				end case;
			else
				dsndouta <= X"00";
			end if;

			if (ch_b = "10") then
				case volb is
					when X"0" =>   dsndoutb <= X"00";
					when X"1" =>   dsndoutb <= X"01";
					when X"2" =>   dsndoutb <= X"02";
					when X"3" =>   dsndoutb <= X"03";
					when X"4" =>   dsndoutb <= X"05";
					when X"5" =>   dsndoutb <= X"07";
					when X"6" =>   dsndoutb <= X"0A";
					when X"7" =>   dsndoutb <= X"0F";
					when X"8" =>   dsndoutb <= X"16";
					when X"9" =>   dsndoutb <= X"1F";
					when X"A" =>   dsndoutb <= X"2C";
					when X"B" =>   dsndoutb <= X"3F";
					when X"C" =>   dsndoutb <= X"5A";
					when X"D" =>   dsndoutb <= X"7F";
					when X"E" =>   dsndoutb <= X"B4";
					when X"F" =>   dsndoutb <= X"FF";
					when others => dsndoutb <= X"00";
				end case;
			elsif (ch_b = "01") then
				case volb is
					when X"0" =>   dsndoutb <= X"00";
					when X"1" =>   dsndoutb <= X"01";
					when X"2" =>   dsndoutb <= X"01";
					when X"3" =>   dsndoutb <= X"02";
					when X"4" =>   dsndoutb <= X"03";
					when X"5" =>   dsndoutb <= X"04";
					when X"6" =>   dsndoutb <= X"06";
					when X"7" =>   dsndoutb <= X"08";
					when X"8" =>   dsndoutb <= X"0B";
					when X"9" =>   dsndoutb <= X"10";
					when X"A" =>   dsndoutb <= X"17";
					when X"B" =>   dsndoutb <= X"20";
					when X"C" =>   dsndoutb <= X"2D";
					when X"D" =>   dsndoutb <= X"40";
					when X"E" =>   dsndoutb <= X"5B";
					when X"F" =>   dsndoutb <= X"FF";
					when others => dsndoutb <= X"00";
				end case;
			else
				dsndoutb <= X"00";
			end if;

			if (ch_c = "10") then
				case volc is
					when X"0" =>   dsndoutc <= X"00";
					when X"1" =>   dsndoutc <= X"01";
					when X"2" =>   dsndoutc <= X"02";
					when X"3" =>   dsndoutc <= X"03";
					when X"4" =>   dsndoutc <= X"05";
					when X"5" =>   dsndoutc <= X"07";
					when X"6" =>   dsndoutc <= X"0A";
					when X"7" =>   dsndoutc <= X"0F";
					when X"8" =>   dsndoutc <= X"16";
					when X"9" =>   dsndoutc <= X"1F";
					when X"A" =>   dsndoutc <= X"2C";
					when X"B" =>   dsndoutc <= X"3F";
					when X"C" =>   dsndoutc <= X"5A";
					when X"D" =>   dsndoutc <= X"7F";
					when X"E" =>   dsndoutc <= X"B4";
					when X"F" =>   dsndoutc <= X"FF";
					when others => dsndoutc <= X"00";
				end case;
			elsif (ch_c = "01") then
				case volc is
					when X"0" =>   dsndoutc <= X"00";
					when X"1" =>   dsndoutc <= X"01";
					when X"2" =>   dsndoutc <= X"01";
					when X"3" =>   dsndoutc <= X"02";
					when X"4" =>   dsndoutc <= X"03";
					when X"5" =>   dsndoutc <= X"04";
					when X"6" =>   dsndoutc <= X"06";
					when X"7" =>   dsndoutc <= X"08";
					when X"8" =>   dsndoutc <= X"0B";
					when X"9" =>   dsndoutc <= X"10";
					when X"A" =>   dsndoutc <= X"17";
					when X"B" =>   dsndoutc <= X"20";
					when X"C" =>   dsndoutc <= X"2D";
					when X"D" =>   dsndoutc <= X"40";
					when X"E" =>   dsndoutc <= X"5B";
					when X"F" =>   dsndoutc <= X"FF";
					when others => dsndoutc <= X"00";
				end case;
			else
				dsndoutc <= X"00";
			end if;

		end if;
	end process;

-- sound mix
	process (CLK,RSTN)
	begin
		if (RSTN = '0') then
			dsndout  <= (others => '0');
			dsndout2 <= (others => '0');
		elsif (CLK'event and CLK = '1') then
			dsndout  <=	("00" & dsndouta) + ("00" & dsndoutb) + ("00" & dsndoutc);
			if (dsndout(9) = '0') then
				dsndout2 <=	dsndout(8 downto 0);
			else
				dsndout2 <=	"111111111";
			end if;
		end if;
	end process;


-- 1bit D/A
	process (CLK,RSTN)
	begin
		if (RSTN = '0') then
			sigma <= (others => '0');
		elsif (CLK'event and CLK = '1') then
			sigma <= (sigma(10) & sigma(10) & '0' & X"00") + sigma + ("00" & dsndout2);
		end if;
	end process;

	process (CLK,RSTN)
	begin
		if (RSTN = '0') then
			sndout_i <= '0';
		elsif (CLK'event and CLK = '1') then
			sndout_i <= sigma(10);
		end if;
	end process;


-- I/O port
	process (CLK,RSTN)
	begin
		if (RSTN = '0') then
			port_ia    <= (others => '1');
			port_ib    <= (others => '1');
			port_ia_f1 <= (others => '1');
			port_ib_f1 <= (others => '1');
		elsif (CLK'event and CLK = '1') then
			port_ia    <= IA;
			port_ib    <= IB;
			port_ia_f1 <= port_ia;
			port_ib_f1 <= port_ib;
		end if;
	end process;


	ENBAN  <= not reg_mixer(6);
	ENBBN  <= not reg_mixer(7);

	OA     <= reg_ioa;
	OB     <= reg_iob;


	SNDOUT <= sndout_i;
	DAO    <= dao_i;


end RTL;
