--
library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use IEEE.std_logic_unsigned.all;

entity CTRLREG is
	port (
		ADD			: in  std_logic_vector(9 downto 0);		-- from CPU
		DATAI		: in  std_logic_vector(7 downto 0);		-- from CPU
		WRN			: in  std_logic;						-- from CPU
		RDN			: in  std_logic;						-- from CPU
		LCDMODE		: in  std_logic;
		MK2MODE		: in  std_logic;
		P66MODE		: in  std_logic;
		FUNCKEY		: in  std_logic_vector(23 downto 0);
		CLK50M		: in  std_logic;
		RSTN		: in  std_logic;
		DATAO		: out std_logic_vector(7 downto 0);		-- to CPU
		LCDCS		: out std_logic;						-- to LCD controller
		LCDSDI		: out std_logic;						-- to LCD controller
		LCDSCL		: out std_logic;						-- to LCD controller
		LCDDONE		: out std_logic;
		CTRLSEL		: out std_logic;
		CTRLRSTN	: out std_logic;
		MEM16K		: out std_logic;
		SC4COLORON	: out std_logic;
		SC4COLORMD	: out std_logic_vector(3 downto 0);
		TAPEACC		: out std_logic;
		TAPEACCH	: out std_logic
	);
end CTRLREG;

architecture RTL of CTRLREG is

	signal reg_lcdadd		: std_logic_vector(5 downto 0);	-- 0x8401(5:0)
	signal reg_lcddat		: std_logic_vector(7 downto 0);	-- 0x8402(7:0)
	signal reg_lcdst		: std_logic;					-- 0x8403(0)
	signal reg_lcdend		: std_logic;					-- 0x8403(0)
	signal reg_lcddone		: std_logic;					-- 0x8404(0)

	signal reg_ctrlselcpu	: std_logic;					-- 0x8410(0)
	signal reg_ctrlsel		: std_logic;					-- 0x8410(1)
	signal reg_ctrlrstcpu	: std_logic;					-- 0x8411(0)

	signal reg_mem16k		: std_logic;					-- 0x8412(0)
	signal reg_sc4coloron	: std_logic;					-- 0x8413(4)
	signal reg_sc4colormd	: std_logic_vector(3 downto 0);	-- 0x8413(3:0)
	signal reg_tapeacc		: std_logic;					-- 0x8414(0)
	signal reg_tapeacch		: std_logic;					-- 0x8414(1)

	signal f_basicrom		: std_logic_vector(87 downto 0);-- 0x8480 - 0x848A
	signal f_cgrom			: std_logic_vector(87 downto 0);-- 0x8490 - 0x849A
	signal f_cg2rom			: std_logic_vector(87 downto 0);-- 0x84A0 - 0x84AA
	signal f_kanjirom		: std_logic_vector(87 downto 0);-- 0x84B0 - 0x84BA
	signal f_voicerom		: std_logic_vector(87 downto 0);-- 0x84C0 - 0x84CA

	signal f_basicrom60		: std_logic_vector(87 downto 0);
	signal f_cgrom60		: std_logic_vector(87 downto 0);
	signal f_cg2rom60		: std_logic_vector(87 downto 0);
	signal f_kanjirom60		: std_logic_vector(87 downto 0);
	signal f_voicerom60		: std_logic_vector(87 downto 0);
	signal f_basicrom62		: std_logic_vector(87 downto 0);
	signal f_cgrom62		: std_logic_vector(87 downto 0);
	signal f_cg2rom62		: std_logic_vector(87 downto 0);
	signal f_kanjirom62		: std_logic_vector(87 downto 0);
	signal f_voicerom62		: std_logic_vector(87 downto 0);
	signal f_basicrom66		: std_logic_vector(87 downto 0);
	signal f_cgrom66		: std_logic_vector(87 downto 0);
	signal f_cg2rom66		: std_logic_vector(87 downto 0);
	signal f_kanjirom66		: std_logic_vector(87 downto 0);
	signal f_voicerom66		: std_logic_vector(87 downto 0);

	signal ctrlsel_i		: std_logic;
	signal ctrlrstn_i		: std_logic;
	signal ctrlrst_f1		: std_logic;
	signal ctrlrstcnt		: std_logic_vector(11 downto 0);

	signal reg_lcdst_f1		: std_logic;
	signal lcdcnt			: std_logic_vector(9 downto 0);
	signal lcddtsel			: std_logic_vector(3 downto 0);

	signal lcdcs_i			: std_logic;
	signal lcdsdi_i			: std_logic;
	signal lcdscl_i			: std_logic;

	signal datao_i			: std_logic_vector(7 downto 0);

	signal rdn_f1			: std_logic;
	signal rdn_f2			: std_logic;
	signal rdn_f3			: std_logic;
	signal wrn_f1			: std_logic;
	signal wrn_f2			: std_logic;
	signal wrn_f3			: std_logic;
	signal add_f1			: std_logic_vector(9 downto 0);
	signal add_f2			: std_logic_vector(9 downto 0);
	signal datai_f1			: std_logic_vector(7 downto 0);
	signal datai_f2			: std_logic_vector(7 downto 0);
	signal rdp				: std_logic;
	signal wrp				: std_logic;

	signal funckey_f1		: std_logic_vector(23 downto 0);
	signal funckey_f2		: std_logic_vector(23 downto 0);
	signal funckey_f3		: std_logic_vector(23 downto 0);

begin

-- CPU I/F
-- data latch
	process (CLK50M,RSTN)
	begin
		if (RSTN = '0') then
			rdn_f1   <= '1';
			rdn_f2   <= '1';
			rdn_f3   <= '1';
			wrn_f1   <= '1';
			wrn_f2   <= '1';
			wrn_f3   <= '1';
			add_f1   <= (others => '0');
			add_f2   <= (others => '0');
			datai_f1 <= (others => '0');
			datai_f2 <= (others => '0');
		elsif (CLK50M'event and CLK50M = '1') then
			rdn_f1   <= RDN;
			rdn_f2   <= rdn_f1;
			rdn_f3   <= rdn_f2;
			wrn_f1   <= WRN;
			wrn_f2   <= wrn_f1;
			wrn_f3   <= wrn_f2;
			add_f1   <= ADD;
			add_f2   <= add_f1;
			datai_f1 <= DATAI;
			datai_f2 <= datai_f1;
		end if;
	end process;

	rdp <= '1' when (rdn_f2 = '0' and rdn_f3 = '1') else '0';
	wrp <= '1' when (wrn_f2 = '0' and wrn_f3 = '1') else '0';

-- register
	process (CLK50M,RSTN)
	begin
		if (RSTN = '0') then
			reg_lcdadd     <= (others => '0');
			reg_lcddat     <= (others => '0');
			reg_lcdst      <= '0';
			reg_lcddone    <= '0';
			reg_ctrlselcpu <= '0';
			reg_ctrlsel    <= '0';
			reg_ctrlrstcpu <= '0';
			reg_mem16k     <= '0';
			reg_sc4coloron <= '1';
			reg_sc4colormd <= "0000";
			reg_tapeacc    <= '1';
			reg_tapeacch   <= '1';
		elsif (CLK50M'event and CLK50M = '1') then
			if (wrp = '1') then
				case add_f2 is
					when "0000000001"=>	reg_lcdadd     <= datai_f2(5 downto 0);
					when "0000000010"=>	reg_lcddat     <= datai_f2;
					when "0000000011"=>	reg_lcdst      <= datai_f2(0);
					when "0000000100"=>	reg_lcddone    <= datai_f2(0);
					when "0000010000"=>	reg_ctrlselcpu <= datai_f2(0);
										reg_ctrlsel    <= datai_f2(1);
					when "0000010001"=>	reg_ctrlrstcpu <= datai_f2(0);
					when "0000010010"=> reg_mem16k     <= datai_f2(0);
					when "0000010011"=> reg_sc4coloron <= datai_f2(4);
										reg_sc4colormd <= datai_f2(3 downto 0);
					when "0000010100"=> reg_tapeacc    <= datai_f2(0);
										reg_tapeacch   <= datai_f2(1);
					when others      => null;
				end case;
			else
				reg_lcdst <= '0';
			end if;
		end if;
	end process;

-- register read
	process (CLK50M,RSTN)
	begin
		if (RSTN = '0') then
			datao_i <= (others => '0');
		elsif (CLK50M'event and CLK50M = '1') then
			if (rdp = '1') then
				case add_f2 is
					when "0000000000"=> datao_i <= "0000000" & LCDMODE;
					when "0000000001"=> datao_i <= "00" & reg_lcdadd;
					when "0000000010"=> datao_i <= reg_lcddat;
					when "0000000011"=> datao_i <= "0000000" & reg_lcdend;
					when "0000000100"=> datao_i <= "0000000" & reg_lcddone;
					when "0000010000"=> datao_i <= "00000" & ctrlsel_i & reg_ctrlsel & reg_ctrlselcpu;
					when "0000010001"=> datao_i <= "0000000" & reg_ctrlrstcpu;
					when "0000010010"=> datao_i <= "0000000" & reg_mem16k;
					when "0000010011"=> datao_i <= "000" & reg_sc4coloron & reg_sc4colormd;
					when "0000010100"=> datao_i <= "000000" & reg_tapeacch & reg_tapeacc;
					when "0000010101"=> datao_i <= "000000" & P66MODE & MK2MODE;
					when "0000100000"=> datao_i <= funckey_f2( 7 downto  0);
					when "0000100001"=> datao_i <= funckey_f2(15 downto  8);
					when "0000100010"=> datao_i <= funckey_f2(23 downto 16);
					when "0010000000"=> datao_i <= f_basicrom(87 downto 80);
					when "0010000001"=> datao_i <= f_basicrom(79 downto 72);
					when "0010000010"=> datao_i <= f_basicrom(71 downto 64);
					when "0010000011"=> datao_i <= f_basicrom(63 downto 56);
					when "0010000100"=> datao_i <= f_basicrom(55 downto 48);
					when "0010000101"=> datao_i <= f_basicrom(47 downto 40);
					when "0010000110"=> datao_i <= f_basicrom(39 downto 32);
					when "0010000111"=> datao_i <= f_basicrom(31 downto 24);
					when "0010001000"=> datao_i <= f_basicrom(23 downto 16);
					when "0010001001"=> datao_i <= f_basicrom(15 downto  8);
					when "0010001010"=> datao_i <= f_basicrom( 7 downto  0);
					when "0010010000"=> datao_i <= f_cgrom(87 downto 80);
					when "0010010001"=> datao_i <= f_cgrom(79 downto 72);
					when "0010010010"=> datao_i <= f_cgrom(71 downto 64);
					when "0010010011"=> datao_i <= f_cgrom(63 downto 56);
					when "0010010100"=> datao_i <= f_cgrom(55 downto 48);
					when "0010010101"=> datao_i <= f_cgrom(47 downto 40);
					when "0010010110"=> datao_i <= f_cgrom(39 downto 32);
					when "0010010111"=> datao_i <= f_cgrom(31 downto 24);
					when "0010011000"=> datao_i <= f_cgrom(23 downto 16);
					when "0010011001"=> datao_i <= f_cgrom(15 downto  8);
					when "0010011010"=> datao_i <= f_cgrom( 7 downto  0);
					when "0010100000"=> datao_i <= f_cg2rom(87 downto 80);
					when "0010100001"=> datao_i <= f_cg2rom(79 downto 72);
					when "0010100010"=> datao_i <= f_cg2rom(71 downto 64);
					when "0010100011"=> datao_i <= f_cg2rom(63 downto 56);
					when "0010100100"=> datao_i <= f_cg2rom(55 downto 48);
					when "0010100101"=> datao_i <= f_cg2rom(47 downto 40);
					when "0010100110"=> datao_i <= f_cg2rom(39 downto 32);
					when "0010100111"=> datao_i <= f_cg2rom(31 downto 24);
					when "0010101000"=> datao_i <= f_cg2rom(23 downto 16);
					when "0010101001"=> datao_i <= f_cg2rom(15 downto  8);
					when "0010101010"=> datao_i <= f_cg2rom( 7 downto  0);
					when "0010110000"=> datao_i <= f_kanjirom(87 downto 80);
					when "0010110001"=> datao_i <= f_kanjirom(79 downto 72);
					when "0010110010"=> datao_i <= f_kanjirom(71 downto 64);
					when "0010110011"=> datao_i <= f_kanjirom(63 downto 56);
					when "0010110100"=> datao_i <= f_kanjirom(55 downto 48);
					when "0010110101"=> datao_i <= f_kanjirom(47 downto 40);
					when "0010110110"=> datao_i <= f_kanjirom(39 downto 32);
					when "0010110111"=> datao_i <= f_kanjirom(31 downto 24);
					when "0010111000"=> datao_i <= f_kanjirom(23 downto 16);
					when "0010111001"=> datao_i <= f_kanjirom(15 downto  8);
					when "0010111010"=> datao_i <= f_kanjirom( 7 downto  0);
					when "0011000000"=> datao_i <= f_voicerom(87 downto 80);
					when "0011000001"=> datao_i <= f_voicerom(79 downto 72);
					when "0011000010"=> datao_i <= f_voicerom(71 downto 64);
					when "0011000011"=> datao_i <= f_voicerom(63 downto 56);
					when "0011000100"=> datao_i <= f_voicerom(55 downto 48);
					when "0011000101"=> datao_i <= f_voicerom(47 downto 40);
					when "0011000110"=> datao_i <= f_voicerom(39 downto 32);
					when "0011000111"=> datao_i <= f_voicerom(31 downto 24);
					when "0011001000"=> datao_i <= f_voicerom(23 downto 16);
					when "0011001001"=> datao_i <= f_voicerom(15 downto  8);
					when "0011001010"=> datao_i <= f_voicerom( 7 downto  0);

					when others      => datao_i <= X"FF";
				end case;
			end if;
		end if;
	end process;

-- PC-6001
	f_basicrom60 <= X"4241534943524F4D363020";	-- BASICROM60
	f_cgrom60    <= X"4347524F4D363020363020";	-- CGROM60 60
	f_cg2rom60   <= X"FFFFFFFFFFFFFFFFFFFFFF";
	f_kanjirom60 <= X"FFFFFFFFFFFFFFFFFFFFFF";
	f_voicerom60 <= X"FFFFFFFFFFFFFFFFFFFFFF";

-- PC-6001mk2
	f_basicrom62 <= X"4241534943524F4D363220";	-- BASICROM62
	f_cgrom62    <= X"4347524F4D363020363220";	-- CGROM60 62
	f_cg2rom62   <= X"4347524F4D36304D363220";	-- CGROM60M62
	f_kanjirom62 <= X"4B414E4A49524F4D363220";	-- KANJIROM62
	f_voicerom62 <= X"564F494345524F4D363220";	-- VOICEROM62

-- PC-6601
	f_basicrom66 <= X"4241534943524F4D363620";	-- BASICROM66
	f_cgrom66    <= X"4347524F4D363020363620";	-- CGROM60 66
	f_cg2rom66   <= X"4347524F4D363620363620";	-- CGROM66 66
	f_kanjirom66 <= X"4B414E4A49524F4D363620";	-- KANJIROM66
	f_voicerom66 <= X"564F494345524F4D363620";	-- VOICEROM66

	f_basicrom <=	f_basicrom60 when (mk2mode = '0' and p66mode = '0') else
					f_basicrom62 when (mk2mode = '1' and p66mode = '0') else
					f_basicrom66 when (mk2mode = '0' and p66mode = '1') else
					(others => '1');

	f_cgrom    <=	f_cgrom60    when (mk2mode = '0' and p66mode = '0') else
					f_cgrom62    when (mk2mode = '1' and p66mode = '0') else
					f_cgrom66    when (mk2mode = '0' and p66mode = '1') else
					(others => '1');

	f_cg2rom   <=	f_cg2rom60   when (mk2mode = '0' and p66mode = '0') else
					f_cg2rom62   when (mk2mode = '1' and p66mode = '0') else
					f_cg2rom66   when (mk2mode = '0' and p66mode = '1') else
					(others => '1');

	f_kanjirom <=	f_kanjirom60 when (mk2mode = '0' and p66mode = '0') else
					f_kanjirom62 when (mk2mode = '1' and p66mode = '0') else
					f_kanjirom66 when (mk2mode = '0' and p66mode = '1') else
					(others => '1');

	f_voicerom <=	f_voicerom60 when (mk2mode = '0' and p66mode = '0') else
					f_voicerom62 when (mk2mode = '1' and p66mode = '0') else
					f_voicerom66 when (mk2mode = '0' and p66mode = '1') else
					(others => '1');


	process (CLK50M,RSTN)
	begin
		if (RSTN = '0') then
			funckey_f1 <= (others => '0');
			funckey_f2 <= (others => '0');
			funckey_f3 <= (others => '0');
		elsif (CLK50M'event and CLK50M = '1') then
			funckey_f1 <= FUNCKEY;
			funckey_f2 <= funckey_f1;
			funckey_f3 <= funckey_f2;
		end if;
	end process;


	process (CLK50M,RSTN)
	begin
		if (RSTN = '0') then
			ctrlsel_i <= '0';
		elsif (CLK50M'event and CLK50M = '1') then
			if (reg_ctrlselcpu = '1') then
				ctrlsel_i <= reg_ctrlsel;
			elsif (funckey_f2(8) = '1' and funckey_f3(8) = '0') then
				ctrlsel_i <= not ctrlsel_i;
			end if;
		end if;
	end process;


	process (CLK50M,RSTN)
	begin
		if (RSTN = '0') then
			ctrlrst_f1 <= '0';
		elsif (CLK50M'event and CLK50M = '1') then
			if (reg_ctrlrstcpu = '1') then
				ctrlrst_f1 <= '1';
			elsif (funckey_f2(11) = '1' and funckey_f3(11) = '0') then
				ctrlrst_f1 <= '1';
			else
				ctrlrst_f1 <= '0';
			end if;
		end if;
	end process;

	process (CLK50M,RSTN)
	begin
		if (RSTN = '0') then
			ctrlrstcnt <= X"FFF";
			ctrlrstn_i <= '0';
		elsif (CLK50M'event and CLK50M = '1') then
			if (ctrlrst_f1 = '1') then
				ctrlrstcnt <= X"000";
				ctrlrstn_i <= '0';
			elsif (ctrlrstcnt = X"FFF") then
				ctrlrstn_i <= '1';
			else
				ctrlrstcnt <= ctrlrstcnt + 1;
				ctrlrstn_i <= '0';
			end if;
		end if;
	end process;


	process (CLK50M,RSTN)
	begin
		if (RSTN = '0') then
			reg_lcdst_f1 <= '0';
		elsif (CLK50M'event and CLK50M = '1') then
			reg_lcdst_f1 <= reg_lcdst;
		end if;
	end process;

	process (CLK50M,RSTN)
	begin
		if (RSTN = '0') then
			lcdcnt <= (others => '1');
		elsif (CLK50M'event and CLK50M = '1') then
			if (reg_lcdst_f1 = '0' and reg_lcdst = '1') then
				lcdcnt <= (others => '0');
			elsif (lcdcnt(9 downto 8) /= "11") then
				lcdcnt <= lcdcnt + 1;
			end if;
		end if;
	end process;

	lcddtsel <= lcdcnt(8 downto 5);

	process (CLK50M,RSTN)
	begin
		if (RSTN = '0') then
			reg_lcdend <= '1';
			lcdcs_i    <= '1';
			lcdsdi_i   <= '1';
			lcdscl_i   <= '1';
		elsif (CLK50M'event and CLK50M = '1') then
			if (reg_lcdst_f1 = '0' and reg_lcdst = '1') then
				reg_lcdend <= '1';
				lcdcs_i    <= '1';
				lcdsdi_i   <= '1';
				lcdscl_i   <= '1';
			else

				if (lcdcnt(9 downto 8) = "11") then
					reg_lcdend <= '0';
				end if;

				lcdcs_i    <= lcdcnt(9);

				if (lcdcnt(9) = '0') then
					lcdscl_i <= lcdcnt(4);
				else
					lcdscl_i <= '1';
				end if;

				if (lcdcnt(9) = '0') then
					case lcddtsel is
						when "0000" => lcdsdi_i <= reg_lcdadd(5);
						when "0001" => lcdsdi_i <= reg_lcdadd(4);
						when "0010" => lcdsdi_i <= reg_lcdadd(3);
						when "0011" => lcdsdi_i <= reg_lcdadd(2);
						when "0100" => lcdsdi_i <= reg_lcdadd(1);
						when "0101" => lcdsdi_i <= reg_lcdadd(0);
						when "0110" => lcdsdi_i <= '1';		-- write(1)/read(0)
						when "0111" => lcdsdi_i <= '0';
						when "1000" => lcdsdi_i <= reg_lcddat(7);
						when "1001" => lcdsdi_i <= reg_lcddat(6);
						when "1010" => lcdsdi_i <= reg_lcddat(5);
						when "1011" => lcdsdi_i <= reg_lcddat(4);
						when "1100" => lcdsdi_i <= reg_lcddat(3);
						when "1101" => lcdsdi_i <= reg_lcddat(2);
						when "1110" => lcdsdi_i <= reg_lcddat(1);
						when "1111" => lcdsdi_i <= reg_lcddat(0);
						when others => lcdsdi_i <= '1';
					end case;
				else
					lcdsdi_i <= '1';
				end if;

			end if;
		end if;
	end process;


	DATAO      <= datao_i;

	LCDCS      <= lcdcs_i;
	LCDSDI     <= lcdsdi_i;
	LCDSCL     <= lcdscl_i;
	LCDDONE    <= reg_lcddone;

	CTRLSEL    <= ctrlsel_i;
	CTRLRSTN   <= ctrlrstn_i;

	MEM16K     <= reg_mem16k;
	SC4COLORON <= reg_sc4coloron;
	SC4COLORMD <= reg_sc4colormd;
	TAPEACC    <= reg_tapeacc;
	TAPEACCH   <= reg_tapeacch;

end RTL;
